import claripy

from .context import get_debugger


def get_dbg_brk_linux64():
    '''
    Return the current brk value in the debugged process (only x86_64 Linux)
    '''
    # TODO this method is so weird, find a unused address to inject code not
    # the base address

    debugger = get_debugger()

    code = b'\x0f\x05'  # syscall

    rax = debugger.get_reg("rax")
    rdi = debugger.get_reg("rdi")
    rip = debugger.get_reg("rip")
    efl = debugger.get_reg("efl")

    debugger.set_reg("rax", 12)  # sys_brk
    debugger.set_reg("rdi", 0)

    base = debugger.image_base()

    inj = base

    save = debugger.get_bytes(inj, len(code))

    debugger.put_bytes(inj, code)

    debugger.set_reg("rip", inj)

    debugger.step_into()
    debugger.wait_ready()

    brk_res = debugger.get_reg("rax")

    debugger.set_reg("rax", rax)
    debugger.set_reg("rdi", rdi)
    debugger.set_reg("rip", rip)
    debugger.set_reg("efl", efl)

    debugger.put_bytes(inj, save)

    return brk_res


def get_dbg_brk_linux32():
    '''
    Return the current brk value in the debugged process (only x86 Linux)
    '''
    # TODO this method is so weird, find a unused address to inject code not
    # the base address

    debugger = get_debugger()

    code = b'\xcd\x80'  # int 0x80

    eax = debugger.get_reg("eax")
    ebx = debugger.get_reg("ebx")
    eip = debugger.get_reg("eip")
    efl = debugger.get_reg("efl")

    debugger.set_reg("eax", 45)  # sys_brk
    debugger.set_reg("ebx", 0)

    base = debugger.image_base()

    inj = base

    save = debugger.get_bytes(inj, len(code))

    debugger.put_bytes(inj, code)

    debugger.set_reg("eip", inj)

    debugger.step_into()
    debugger.wait_ready()

    brk_res = debugger.get_reg("eax")

    debugger.set_reg("eax", eax)
    debugger.set_reg("ebx", ebx)
    debugger.set_reg("eip", eip)
    debugger.set_reg("efl", efl)

    debugger.put_bytes(inj, save)

    return brk_res


def get_linux_brk(bits):
    if bits == 64:
        curr_brk = get_dbg_brk_linux64()
        #print "get_linux_brk: current brk = 0x%x" % curr_brk
        return claripy.BVV(curr_brk, 64)
    else:
        curr_brk = get_dbg_brk_linux32()
        #print "get_linux_brk: current brk = 0x%x" % curr_brk
        return claripy.BVV(curr_brk, 32)



from .memory import SimSymbolicDbgMemory
from .context import load_project, get_memory_type, set_memory_type, get_debugger, SIMPROCS_FROM_CLE, ONLY_GOT_FROM_CLE, GET_ALL_DISCARD_CLE
from .brk import get_linux_brk
from .got_builder import *
from .idata_builder import *

import angr
import claripy

import logging
l = logging.getLogger("angrdbg.core")

import sys
if sys.version_info >= (3, 0):
    long = int
else:
    bytes = str


def get_logger():
    return logging.getLogger("angrdbg")


def get_registers():
    project = load_project()
    regs = {}
    
    for reg in sorted(project.arch.registers,
                      key=lambda x: project.arch.registers.get(x)[1]):
        if reg in ("sp", "bp", "ip"):
            continue
        try:
            regs[reg] = debugger.get_reg(reg)
        except BaseException:
            pass
    
    return regs

def StateShot(regs={}, sync_brk=True, check_dbg=False, **kwargs):
    debugger = get_debugger()

    if not check_dbg:
        if not debugger.is_active():
            raise RuntimeError(
                "The debugger must be active and suspended before calling StateShot")
        debugger.refresh_memory()

    project = load_project()

    debugger.before_stateshot()

    mem = SimSymbolicDbgMemory(
        memory_backer=project.loader.memory,
        permissions_backer=None,
        memory_id="mem")

    state = project.factory.blank_state(plugins={"memory": mem}, **kwargs)

    for reg in sorted(project.arch.registers,
                      key=lambda x: project.arch.registers.get(x)[1]):
        if reg in ("sp", "bp", "ip", "pc"):
            continue
        try:
            if reg in regs:
                setattr(state.regs, reg, regs[reg])
            else:
                setattr(state.regs, reg, debugger.get_reg(reg))
            #print(reg, getattr(state.regs, reg), debugger.get_reg(reg))
        except BaseException:
            pass

    if project.simos.name == "Linux":
        # inject code to get brk if we are on linux x86/x86_64
        if sync_brk and project.arch.name in ("AMD64", "X86"):
            try:
                state.posix.set_brk(get_linux_brk(project.arch.bits))
            except angr.errors.SimMemoryError:
                l.warning("failed to sychronize brk")

        if get_memory_type() == SIMPROCS_FROM_CLE:
            # insert simprocs when possible or resolve the symbol
            state = build_mixed_got(project, state)
        elif get_memory_type() == ONLY_GOT_FROM_CLE:
            # load the entire got from cle with stubs
            state = build_cle_got(project, state)
        elif get_memory_type() == GET_ALL_DISCARD_CLE:
            # angr must not execute loader code so all symbols must be resolved
            state = build_bind_now_got(project, state)
    
    elif project.simos.name == "Win32":
        if get_memory_type() == SIMPROCS_FROM_CLE:
            # insert simprocs when possible or resolve the symbol
            state = build_mixed_idata(project, state)
        elif get_memory_type() == ONLY_GOT_FROM_CLE:
            # load the entire idata from cle with stubs
            state = build_cle_idata(project, state)
    
    debugger.after_stateshot(state)
    
    return state


class StateManager(object):
    def __init__(self, state=None, **kwargs):
        self.state = StateShot(**kwargs) if state is None else state
        self.symbolics = {}
        self.debugger = get_debugger()

    def sim(self, key, size=None):
        '''
        key: memory address(int) or register name(str)
        size: size of object in bytes
        '''
        project = load_project()
        if key in project.arch.registers:
            if size is None:
                size = project.arch.registers[key][1]
            size *= 8
            s = claripy.BVS("angrdbg_reg_" + str(key), size)
            setattr(self.state.regs, key, s)
            self.symbolics[key] = (s, size)
        elif isinstance(key, int) or isinstance(key, long):
            if size is None:
                size = project.arch.bits
            else:
                size *= 8
            s = claripy.BVS("angrdbg_mem_" + hex(key), size)
            self.state.memory.store(key, s)
            self.symbolics[key] = (s, size)
        elif isinstance(key, claripy.ast.bv.BV):
            key = self.state.solver.eval(key, cast_to=int)
            self.sim(key, size)
        else:
            raise ValueError(
                "key must be a register name or a memory address, not %s" % str(
                    type(key)))
        return key

    def sim_from_set(self, simset):
        for key in simset.symbolics:
            if key in load_project().arch.registers:
                setattr(self.state.regs, key, simset.symbolics[key][0])
            else:
                self.state.memory.store(key, simset.symbolics[key][0])

    def __getitem__(self, key):
        if key in load_project().arch.registers:
            return getattr(self.state.regs, key)
        elif isinstance(key, int) or isinstance(key, long):
            return self.state.mem[key]
        elif isinstance(key, claripy.ast.bv.BV):
            return self.state.mem[key]
        else:
            raise ValueError("key must be a register name or a memory address")

    def __setitem__(self, key, value):
        if key in load_project().arch.registers:
            setattr(self.state.regs, key, value)
        elif isinstance(key, int) or isinstance(key, long) or isinstance(key, claripy.ast.bv.BV):
            self.state.memory[key] = value
        else:
            raise ValueError("key must be a register name or a memory address")

    def simulation_manager(self):
        return load_project().factory.simulation_manager(self.state)

    def get_symbolic(self, key):
        return self.symbolics.get(key)

    def get_state(self):
        return self.state

    def to_dbg(self, found_state):
        if isinstance(found_state, StateManager):
            return self.to_dbg(found_state.state)
        for key in self.symbolics:
            try:
                if key in load_project().arch.registers:
                    r = found_state.solver.eval(
                        self.symbolics[key][0], cast_to=int)
                    self.debugger.set_reg(key, r)
                else:
                    r = found_state.solver.eval(
                        self.symbolics[key][0], cast_to=bytes)
                    self.debugger.put_bytes(key, r)
            except Exception as ee:
                l.warning(" >> failed to write %s to debugger" % key)
                #print ee

    def concretize(self, found_state):
        if isinstance(found_state, StateManager):
            return self.concretize(found_state.state)
        ret = {}
        for key in self.symbolics:
            try:
                if key in load_project().arch.registers:
                    r = found_state.solver.eval(
                        self.symbolics[key][0], cast_to=int)
                    ret[key] = r
                else:
                    r = found_state.solver.eval(
                        self.symbolics[key][0], cast_to=bytes)
                    ret[key] = r
            except Exception as ee:
                l.warning(" >> failed to concretize %s" % key)
                #print ee
        return ret

    def print_symbolics(self):
        for key in self.symbolics:
            k = key
            if isinstance(key, int) or isinstance(key, long):
                k = "0x%x" % key
            print ("%s ==> %s" %(str(k), str(self.symbolics[key])))



#!/usr/bin/env python
"""
classic rpyc server running a SlaveService + angrdbg + IPython shell
usage:
    angrdbg-srv.py                            # default settings
    angrdbg-srv.py --host HOST --port PORT    # custom settings

    # ssl-authenticated server (keyfile and certfile are required)
    angrdbg-srv.py --ssl-keyfile keyfile.pem --ssl-certfile certfile.pem --ssl-cafile cafile.pem
"""
import sys
import os
import rpyc
import threading
import signal
import Queue

from plumbum import cli
from rpyc.utils.server import Server
from rpyc.utils.classic import DEFAULT_SERVER_PORT, DEFAULT_SERVER_SSL_PORT
from rpyc.utils.registry import REGISTRY_PORT
from rpyc.utils.registry import UDPRegistryClient, TCPRegistryClient
from rpyc.utils.authenticators import SSLAuthenticator
from rpyc.lib import setup_logger
from rpyc.core import SlaveService

BANNER = "[angrdbg server v1.0]"

#######################
import angr
import claripy
import pyvex
import angrdbg
import IPython
#from angrdbg import *
#######################


class WeirdServer(Server):  # n1 threaded n2 forked

    def __init__(self, service, done_event, **kwargs):
        self.num_conns = 2
        self.thread = None
        self.proc = None
        self.done_event = done_event
        Server.__init__(self, service, **kwargs)

    @classmethod
    def _handle_sigchld(cls, signum, unused):
        try:
            while True:
                pid, dummy = os.waitpid(-1, os.WNOHANG)
                if pid <= 0:
                    break
        except OSError:
            pass
        # re-register signal handler (see man signal(2), under Portability)
        signal.signal(signal.SIGCHLD, cls._handle_sigchld)

    def _accept_method(self, sock):
        self.num_conns -= 1

        if self.num_conns == 1:
            t = threading.Thread(
                target=self._authenticate_and_serve_client,
                args=[sock])
            t.start()

            self.thread = t
        else:
            pid = os.fork()
            if pid == 0:
                # child
                try:
                    self.logger.debug("child process created")
                    # 76: call signal.siginterrupt(False) in forked child
                    signal.siginterrupt(signal.SIGCHLD, False)
                    self.listener.close()
                    self.clients.clear()
                    self._authenticate_and_serve_client(sock)
                except BaseException:
                    self.logger.exception(
                        "child process terminated abnormally")
                else:
                    self.logger.debug("child process terminated")
                finally:
                    self.logger.debug("child terminated")
                    os._exit(0)
            else:
                # parent
                self.proc = pid
                sock.close()

        if self.num_conns == 0:
            self.done_event.set()
            self.listener.close()
            self.join()

    def join(self):
        self.thread.join()

        try:
            pid, dummy = os.waitpid(self.proc, 0)  # os.WNOHANG)
        except OSError as ee:
            print ee


class AngrDbgServer(cli.Application):
    port = cli.SwitchAttr(["-p", "--port"], cli.Range(0, 65535), default=None,
                          help="The TCP listener port (default = %s, default for SSL = %s)" %
                          (DEFAULT_SERVER_PORT, DEFAULT_SERVER_SSL_PORT), group="Socket Options")
    host = cli.SwitchAttr(
        ["--host"],
        str,
        default="127.0.0.1",
        help="The host to bind to. "
        "The default is INADDR_ANY",
        group="Socket Options")
    ipv6 = cli.Flag(["--ipv6"], help="Enable IPv6", group="Socket Options")

    logfile = cli.SwitchAttr(
        "--logfile",
        str,
        default=None,
        help="Specify the log file to use; "
        "the default is stderr",
        group="Logging")
    quiet = cli.Flag(["-q",
                      "--quiet"],
                     help="Quiet mode (only errors will be logged)",
                     group="Logging")

    ssl_keyfile = cli.SwitchAttr(
        "--ssl-keyfile",
        cli.ExistingFile,
        help="The keyfile to use for SSL. Required for SSL",
        group="SSL",
        requires=["--ssl-certfile"])
    ssl_certfile = cli.SwitchAttr(
        "--ssl-certfile",
        cli.ExistingFile,
        help="The certificate file to use for SSL. Required for SSL",
        group="SSL",
        requires=["--ssl-keyfile"])
    ssl_cafile = cli.SwitchAttr(
        "--ssl-cafile",
        cli.ExistingFile,
        help="The certificate authority chain file to use for SSL. Optional; enables client-side "
        "authentication",
        group="SSL",
        requires=["--ssl-keyfile"])

    auto_register = cli.Flag(
        "--register",
        help="Asks the server to attempt registering with "
        "a registry server. By default, the server will not attempt to register",
        group="Registry")
    registry_type = cli.SwitchAttr(
        "--registry-type",
        cli.Set(
            "UDP",
            "TCP"),
        default="UDP",
        help="Specify a UDP or TCP registry",
        group="Registry")
    registry_port = cli.SwitchAttr(
        "--registry-port",
        cli.Range(
            0,
            65535),
        default=REGISTRY_PORT,
        help="The registry's UDP/TCP port",
        group="Registry")
    registry_host = cli.SwitchAttr(
        "--registry-host",
        str,
        default=None,
        help="The registry host machine. For UDP, the default is 255.255.255.255; "
        "for TCP, a value is required",
        group="Registry")

    def main(self):
        if self.registry_type == "UDP":
            if self.registry_host is None:
                self.registry_host = "255.255.255.255"
            self.registrar = UDPRegistryClient(
                ip=self.registry_host, port=self.registry_port)
        else:
            if self.registry_host is None:
                raise ValueError(
                    "With TCP registry, you must specify --registry-host")
            self.registrar = TCPRegistryClient(
                ip=self.registry_host, port=self.registry_port)

        if self.ssl_keyfile:
            self.authenticator = SSLAuthenticator(
                self.ssl_keyfile, self.ssl_certfile, self.ssl_cafile)
            default_port = DEFAULT_SERVER_SSL_PORT
        else:
            self.authenticator = None
            default_port = DEFAULT_SERVER_PORT
        if self.port is None:
            self.port = default_port

        setup_logger(self.quiet, self.logfile)

        sys.stdout.write(
            BANNER + " starting at %s %s\n" %
            (self.host, self.port))
        sys.stdout.flush()

        done_event = threading.Event()
        srv = WeirdServer(
            SlaveService,
            done_event,
            hostname=self.host,
            port=self.port,
            reuse_addr=True,
            ipv6=self.ipv6,
            authenticator=self.authenticator,
            registrar=self.registrar,
            auto_register=self.auto_register)

        t = threading.Thread(target=self._serve, args=[srv])
        t.start()

        # wait for 2 connections
        done_event.wait()

        IPython.embed(
            banner1=BANNER + " client connected\n",
            banner2="",  # "tip: call serve_all() on the client to have a full working shell here.",
            exit_msg=BANNER + " shell closed.\nexiting...\n"
        )

        os.kill(srv.proc, signal.SIGKILL)
        os._exit(0)

    def _serve(self, srv):
        srv.start()

        sys.stdout.write("\n" + BANNER + " client disconnected.\nexiting...\n")
        os._exit(0)


def main():
    AngrDbgServer.run()


'''simple client
import rpyc
import thread

conn1 = rpyc.classic.connect("localhost")
conn2 = rpyc.classic.connect("localhost")
thread.start_new_thread(conn2.serve_all, tuple())

'''



from .context import get_debugger

import logging

l = logging.getLogger("angrdbg.got_builder")

#py2 and 3 support
try:
    xrange
    import itertools
    def range(start, stop, step):
        return iter(itertools.count(start, step).next, stop)
except NameError:
    pass

'''
def get_other_symbols_addrs(proj):
    i = 1
    while True:
        try:
            sym = proj.loader.main_object.get_symbol(i)
        except: break
        if sym.rebased_addr > 0:
            yield sym.name, sym.relative_addr
        i += 1
'''

def build_cle_got(proj, state):
    debugger = get_debugger()

    try:
        got_start, got_end = debugger.get_got()
    except BaseException:
        l.warning("cannot find .got.plt section, build_cle_got failed")
        return state

    entry_len = proj.arch.bits // 8

    got_start += 3 * entry_len  # skip first 3 entries

    '''
    print "## angr got - before ##"
    for a in range(got_start, got_end, entry_len):
        print "0x%x:  0x%x" % (a, state.solver.eval(getattr(state.mem[a], "uint%d_t" % proj.arch.bits).resolved))
    print
    '''

    empty_state = proj.factory.blank_state()
    state.memory.store(
        got_start,
        empty_state.memory.load(
            got_start,
            got_end -
            got_start))

    '''
    print "## angr got - final ##"
    for a in range(got_start, got_end, entry_len):
        print "0x%x:  0x%x" % (a, state.solver.eval(getattr(state.mem[a], "uint%d_t" % proj.arch.bits).resolved))
    print
    '''

    return state


def build_mixed_got(proj, state):
    debugger = get_debugger()

    try:
        got_start, got_end = debugger.get_got()
    except BaseException:
        l.warning("cannot find .got.plt section, build_mixed_got failed")
        return state

    try:
        plt_start, plt_end = debugger.get_plt()
    except BaseException:
        l.warning("cannot find .plt section, build_mixed_got failed")
        return state

    entry_len = proj.arch.bits // 8
    get_mem = debugger.get_dword if entry_len == 4 else debugger.get_qword

    got_start += 3 * entry_len  # skip first 3 entries

    '''
    print "## angr got - before ##"
    for a in range(got_start, got_end, entry_len):
        print "0x%x:  0x%x" % (a, state.solver.eval(getattr(state.mem[a], "uint%d_t" % proj.arch.bits).resolved))
    print
    '''

    empty_state = proj.factory.blank_state()

    for a in range(got_start, got_end, entry_len):
        state_val = empty_state.solver.eval(
            getattr(
                empty_state.mem[a],
                "uint%d_t" %
                proj.arch.bits).resolved)
        if state_val in proj._sim_procedures:
            if proj._sim_procedures[state_val].is_stub:  # real simprocs or not?
                dbg_val = get_mem(a)
                name = proj._sim_procedures[state_val].display_name

                if dbg_val >= plt_end or dbg_val < plt_start:  # already resolved by the loader in the dbg
                    setattr(state.mem[a], "uint%d_t" % proj.arch.bits, dbg_val)
                else:
                    ea = debugger.resolve_name(name)
                    if ea is not None:
                        setattr(state.mem[a], "uint%d_t" % proj.arch.bits, ea)
            else:
                setattr(state.mem[a], "uint%d_t" % proj.arch.bits, state_val)
    
    
    '''
    print "## angr got - final ##"
    for a in range(got_start, got_end, entry_len):
        print "0x%x:  0x%x" % (a, state.solver.eval(getattr(state.mem[a], "uint%d_t" % proj.arch.bits).resolved))
    print
    '''

    return state


def build_bind_now_got(proj, state):
    debugger = get_debugger()

    try:
        got_start, got_end = debugger.get_got()
    except BaseException:
        l.warning("cannot find .got.plt section, build_bind_now_got failed")
        return state

    try:
        plt_start, plt_end = debugger.get_plt()
    except BaseException:
        l.warning("cannot find .plt section, build_bind_now_got failed")
        return state

    entry_len = proj.arch.bits // 8
    get_mem = debugger.get_dword if entry_len == 4 else debugger.get_qword

    got_start += 3 * entry_len  # skip first 3 entries

    '''
    print "## angr got - before ##"
    for a in range(got_start, got_end, entry_len):
        print "0x%x:  0x%x" % (a, state.solver.eval(getattr(state.mem[a], "uint%d_t" % proj.arch.bits).resolved))
    print
    '''

    empty_state = proj.factory.blank_state()

    for a in range(got_start, got_end, entry_len):
        state_val = empty_state.solver.eval(
            getattr(
                empty_state.mem[a],
                "uint%d_t" %
                proj.arch.bits).resolved)
        if state_val in proj._sim_procedures:
            dbg_val = get_mem(a)
            name = proj._sim_procedures[state_val].display_name

            if dbg_val >= plt_end or dbg_val < plt_start:  # already resolved by the loader in the dbg
                setattr(state.mem[a], "uint%d_t" % proj.arch.bits, dbg_val)
            else:
                ea = debugger.resolve_name(name)
                if ea is not None:
                    setattr(state.mem[a], "uint%d_t" % proj.arch.bits, ea)
    
    '''
    print "## angr got - final ##"
    for a in range(got_start, got_end, entry_len):
        print "0x%x:  0x%x" % (a, state.solver.eval(getattr(state.mem[a], "uint%d_t" % proj.arch.bits).resolved))
    print
    '''

    return state



import angr
import logging

l = logging.getLogger("angrdbg.context")

from .abstract_debugger import Debugger

import sys
if sys.version_info >= (3, 0):
    long = int
    _MAIN_OPTS=lambda debugger: { 'base_addr': debugger.image_base() , 'force_rebase': True}
else:
    _MAIN_OPTS=lambda debugger: { 'custom_base_addr': debugger.image_base() }


project = None
debugger = Debugger()


def reload_project(input_file=None):
    global project
    l.info("creating angr project...")
    if input_file is None:
        input_file = debugger.input_file()
    project = angr.Project(input_file,
                            main_opts=_MAIN_OPTS(debugger),
                            load_options={ "auto_load_libs": False })
    l.info("angr project created.")
    return project

def load_project(input_file=None):
    global project
    if project == None:
        return reload_project(input_file)
    return project

SIMPROCS_FROM_CLE = 0
ONLY_GOT_FROM_CLE = 1
USE_CLE_MEMORY = 2
GET_ALL_DISCARD_CLE = 3

memory_types = {
    "SIMPROCS_FROM_CLE": SIMPROCS_FROM_CLE,
    "ONLY_GOT_FROM_CLE" : ONLY_GOT_FROM_CLE,
    "USE_CLE_MEMORY" : USE_CLE_MEMORY,
    "GET_ALL_DISCARD_CLE" : GET_ALL_DISCARD_CLE
}

memory_type = SIMPROCS_FROM_CLE

def set_memory_type(value):
    global memory_type
    if value not in range(0,4):
        raise ValueError("invalid memory_type")
    memory_type = value

def get_memory_type():
    global memory_type
    return memory_type

def register_debugger(dbginstance):
    global debugger
    debugger = dbginstance

def get_debugger():
    global debugger
    return debugger



from .context import get_debugger

import logging

l = logging.getLogger("angrdbg.idata_builder")

#py2 and 3 support
try:
    xrange
    import itertools
    def range(start, stop, step):
        return iter(itertools.count(start, step).next, stop)
except NameError:
    pass


def build_cle_idata(proj, state):
    debugger = get_debugger()

    try:
        idata_start, idata_end = debugger.get_idata()
    except BaseException:
        l.warning("cannot find .idata section, build_cle_idata failed")
        return state

    entry_len = proj.arch.bits // 8
    
    empty_state = proj.factory.blank_state()
    state.memory.store(
        idata_start,
        empty_state.memory.load(
            idata_start,
            idata_end -
            idata_start))

    return state


def build_mixed_idata(proj, state):
    debugger = get_debugger()

    try:
        idata_start, idata_end = debugger.get_idata()
    except BaseException:
        l.warning("cannot find .idata section, build_mixed_idata failed")
        return state

    entry_len = proj.arch.bits // 8
    get_mem = debugger.get_dword if entry_len == 4 else debugger.get_qword

    empty_state = proj.factory.blank_state()

    for a in range(idata_start, idata_end, entry_len):
        state_val = empty_state.solver.eval(
            getattr(
                empty_state.mem[a],
                "uint%d_t" %
                proj.arch.bits).resolved)
        if state_val in proj._sim_procedures:
            if proj._sim_procedures[state_val].is_stub:  # real simprocs or not?
                dbg_val = get_mem(a)
                setattr(state.mem[a], "uint%d_t" % proj.arch.bits, dbg_val)
            else:
                setattr(state.mem[a], "uint%d_t" % proj.arch.bits, state_val)
                
    return state


